@@ -1,11 +1,17 @@
+use std::cell::Ref;
+use std::cell::RefCell;
+use std::cell::RefMut;
+
 use self::config::BuiltConfig;
 pub use self::config::Config;
 use super::util::{
     self, define_es_module, define_property, has_use_strict, initialize_to_undefined,
     local_name_for_src, make_descriptor, make_require_call, use_strict, Exports, ModulePass, Scope,
 };
+use crate::path::{ImportResolver, NoopImportResolver};
 use fxhash::FxHashSet;
 use swc_atoms::js_word;
+use swc_common::FileName;
 use swc_common::{sync::Lrc, Mark, SourceMap, DUMMY_SP};
 use swc_ecma_ast::*;
 use swc_ecma_transforms_base::helper;
@@ -25,21 +31,54 @@ pub fn umd(cm: Lrc<SourceMap>, root_mark: Mark, config: Config) -> impl Fold {
         cm,
 
         in_top_level: Default::default(),
-        scope: Default::default(),
+        scope: RefCell::new(Default::default()),
         exports: Default::default(),
+
+        resolver: None::<(NoopImportResolver, _)>,
     }
 }
 
-struct Umd {
+pub fn umd_with_resolver<R>(
+    resolver: R,
+    base: FileName,
+    cm: Lrc<SourceMap>,
+    root_mark: Mark,
+    config: Config,
+) -> impl Fold
+where
+    R: ImportResolver,
+{
+    Umd {
+        config: config.build(cm.clone()),
+        root_mark,
+        cm,
+
+        in_top_level: Default::default(),
+        scope: Default::default(),
+        exports: Default::default(),
+
+        resolver: Some((resolver, base)),
+    }
+}
+
+struct Umd<R>
+where
+    R: ImportResolver,
+{
     cm: Lrc<SourceMap>,
     root_mark: Mark,
     in_top_level: bool,
     config: BuiltConfig,
-    scope: Scope,
+    scope: RefCell<Scope>,
     exports: Exports,
+
+    resolver: Option<(R, FileName)>,
 }
 
-impl Fold for Umd {
+impl<R> Fold for Umd<R>
+where
+    R: ImportResolver,
+{
     noop_fold_type!();
 
     fn fold_expr(&mut self, expr: Expr) -> Expr {
@@ -81,13 +120,15 @@ impl Fold for Umd {
             };
 
             match decl {
-                ModuleDecl::Import(import) => self.scope.insert_import(import),
+                ModuleDecl::Import(import) => self.scope.borrow_mut().insert_import(import),
 
                 ModuleDecl::ExportAll(..)
                 | ModuleDecl::ExportDecl(..)
                 | ModuleDecl::ExportDefaultDecl(..)
                 | ModuleDecl::ExportDefaultExpr(..)
                 | ModuleDecl::ExportNamed(..) => {
+                    let mut scope_ref_mut = self.scope.borrow_mut();
+                    let scope = &mut *scope_ref_mut;
                     has_export = true;
                     if !self.config.config.strict && !emitted_esmodule {
                         emitted_esmodule = true;
@@ -118,7 +159,7 @@ impl Fold for Umd {
                         }) => {}
 
                         ModuleDecl::ExportAll(ref export) => {
-                            self.scope
+                            scope
                                 .import_types
                                 .entry(export.src.value.clone())
                                 .and_modify(|v| *v = true);
@@ -131,6 +172,8 @@ impl Fold for Umd {
                         }
                         _ => {}
                     }
+                    drop(scope);
+                    drop(scope_ref_mut);
 
                     match decl {
                         ModuleDecl::ExportAll(export) => export_alls.push(export),
@@ -176,10 +219,11 @@ impl Fold for Umd {
                         }) => {
                             extra_stmts.push(Stmt::Decl(Decl::Var(var.clone().fold_with(self))));
 
+                            let scope = &mut *self.scope.borrow_mut();
                             var.decls.visit_with(
                                 &Invalid { span: DUMMY_SP } as _,
                                 &mut VarCollector {
-                                    to: &mut self.scope.declared_vars,
+                                    to: &mut scope.declared_vars,
                                 },
                             );
 
@@ -189,7 +233,7 @@ impl Fold for Umd {
                                 decl.visit_with(&Invalid { span: DUMMY_SP } as _, &mut v);
 
                                 for ident in found.drain(..) {
-                                    self.scope
+                                    scope
                                         .exported_vars
                                         .entry((ident.sym.clone(), ident.span.ctxt()))
                                         .or_default()
@@ -293,10 +337,13 @@ impl Fold for Umd {
 
                         // export { foo } from 'foo';
                         ModuleDecl::ExportNamed(export) => {
+                            let mut scope_ref_mut = self.scope.borrow_mut();
+                            let scope = &mut *scope_ref_mut;
                             let imported = export.src.clone().map(|src| {
-                                self.scope
-                                    .import_to_export(&src, !export.specifiers.is_empty())
+                                scope.import_to_export(&src, !export.specifiers.is_empty())
                             });
+                            drop(scope);
+                            drop(scope_ref_mut);
 
                             stmts.reserve(export.specifiers.len());
 
@@ -316,29 +363,29 @@ impl Fold for Umd {
                                 let is_import_default = orig.sym == js_word!("default");
 
                                 let key = (orig.sym.clone(), orig.span.ctxt());
-                                if self.scope.declared_vars.contains(&key) {
-                                    self.scope
-                                        .exported_vars
-                                        .entry(key.clone())
-                                        .or_default()
-                                        .push(
-                                            exported
-                                                .clone()
-                                                .map(|i| (i.sym.clone(), i.span.ctxt()))
-                                                .unwrap_or_else(|| {
-                                                    (orig.sym.clone(), orig.span.ctxt())
-                                                }),
-                                        );
+                                let mut scope_ref_mut = self.scope.borrow_mut();
+                                let scope = &mut *scope_ref_mut;
+                                if scope.declared_vars.contains(&key) {
+                                    scope.exported_vars.entry(key.clone()).or_default().push(
+                                        exported
+                                            .clone()
+                                            .map(|i| (i.sym.clone(), i.span.ctxt()))
+                                            .unwrap_or_else(|| {
+                                                (orig.sym.clone(), orig.span.ctxt())
+                                            }),
+                                    );
                                 }
 
                                 if let Some(ref src) = export.src {
                                     if is_import_default {
-                                        self.scope
+                                        scope
                                             .import_types
                                             .entry(src.value.clone())
                                             .or_insert(false);
                                     }
                                 }
+                                drop(scope);
+                                drop(scope_ref_mut);
 
                                 let value = match imported {
                                     Some(ref imported) => Box::new(
@@ -415,7 +462,8 @@ impl Fold for Umd {
             elems: vec![],
         };
 
-        let mut factory_params = Vec::with_capacity(self.scope.imports.len() + 1);
+        let scope = &mut *self.scope.borrow_mut();
+        let mut factory_params = Vec::with_capacity(scope.imports.len() + 1);
         let mut factory_args = Vec::with_capacity(factory_params.capacity());
         let mut global_factory_args = Vec::with_capacity(factory_params.capacity());
         if has_export {
@@ -474,7 +522,7 @@ impl Fold for Umd {
         };
 
         for export in export_alls {
-            stmts.push(self.scope.handle_export_all(
+            stmts.push(scope.handle_export_all(
                 exports_ident.clone(),
                 exported_names.clone(),
                 export,
@@ -485,7 +533,7 @@ impl Fold for Umd {
             stmts.push(initialize_to_undefined(exports_ident, initialized).into_stmt());
         }
 
-        for (src, import) in self.scope.imports.drain(..) {
+        for (src, import) in scope.imports.drain(..) {
             let global_ident = Ident::new(self.config.global_name(&src), DUMMY_SP);
             let import = import.unwrap_or_else(|| {
                 (
@@ -503,12 +551,13 @@ impl Fold for Umd {
                 decorators: Default::default(),
                 pat: Pat::Ident(ident.clone().into()),
             });
-            factory_args.push(make_require_call(self.root_mark, src.clone()).as_arg());
+            factory_args
+                .push(make_require_call(&self.resolver, self.root_mark, src.clone()).as_arg());
             global_factory_args.push(quote_ident!("global").make_member(global_ident).as_arg());
 
             {
                 // handle interop
-                let ty = self.scope.import_types.get(&src);
+                let ty = scope.import_types.get(&src);
 
                 match ty {
                     Some(&wildcard) => {
@@ -740,7 +789,7 @@ impl Fold for Umd {
             var.decls.visit_with(
                 &Invalid { span: DUMMY_SP } as _,
                 &mut VarCollector {
-                    to: &mut self.scope.declared_vars,
+                    to: &mut self.scope.borrow_mut().declared_vars,
                 },
             );
         }
@@ -754,17 +803,20 @@ impl Fold for Umd {
     mark_as_nested!();
 }
 
-impl ModulePass for Umd {
+impl<R> ModulePass for Umd<R>
+where
+    R: ImportResolver,
+{
     fn config(&self) -> &util::Config {
         &self.config.config
     }
 
-    fn scope(&self) -> &Scope {
-        &self.scope
+    fn scope(&self) -> Ref<Scope> {
+        self.scope.borrow()
     }
 
-    fn scope_mut(&mut self) -> &mut Scope {
-        &mut self.scope
+    fn scope_mut(&mut self) -> RefMut<Scope> {
+        self.scope.borrow_mut()
     }
 
     /// ```js
